import unittest
import pandas as pd
import numpy as np
import tempfile
import os

from VESIcal import sample_class, batchfile, calculate_classes, utils

class MockCalculate(calculate_classes.Calculate):
    def __init__(self, result):
        self.result = result

class TestSaveResults(unittest.TestCase):

    def setUp(self):
        # scalar
        self.scalar = 42
        self.np_scalar = np.float64(3.14159)
        self.calc_scalar = MockCalculate(result=42)
        
        # dict
        self.dict_obj = {"SiO2": 50.2, "MgO": 7.5}
        self.calc_dict = MockCalculate(result={"A": 1, "B": 2})
        
        # tuple
        self.tup_obj = (44, 45, 46)
        
        # pandas objects
        self.df = pd.DataFrame({"A": [1, 2], "B": [3, 4]})
        self.series = pd.Series(self.dict_obj)
        self.calc_df = MockCalculate(result=pd.DataFrame({"col": [10, 20]}))
        self.calc_series = MockCalculate(result=pd.Series({"Z": 99}))
        
        # VESIcal objects
        # create temporary place to store written files
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path = self.temp_dir.name
        
        # Create a temporary CSV file to import to BatchFile object
        df = pd.DataFrame({
            'Lable': ['Test1'],
            'SiO2': [72.4],
            'Al2O3': [13.1],
            'FeO': [2.4]
        })
        with tempfile.NamedTemporaryFile(mode='w', suffix='.csv', delete=False) as f:
            df.to_csv(f.name, index=False)
            temp_filename = f.name        
        
        # create objects of supported types for testing
        # VESIcal objects
        self.sample = sample_class.Sample({"SiO2": 70, "Al2O3": 15})
        self.batch = batchfile.BatchFile(temp_filename, label="Sample")
        self.calc_batch = MockCalculate(result=self.batch)
        self.calc_sample = MockCalculate(result=self.sample)

        os.remove(temp_filename)
                
        # mixed list
        self.mixed_data = [
            self.scalar,
            self.np_scalar,
            self.dict_obj,
            self.tup_obj,
            self.df,
            self.series,
            self.batch,
            self.sample,
            self.calc_scalar,
            self.calc_dict,
            self.calc_df,
            self.calc_series,
            self.calc_batch,
            self.calc_sample,
        ]

    def tearDown(self):
        self.temp_dir.cleanup()
        
    def test_save_csv_single_sheet(self):
        with tempfile.TemporaryDirectory() as tmp:
            path = os.path.join(tmp, "output.csv")
            utils.save_results(path, self.mixed_data, filetype="csv", mode="single_sheet")
            self.assertTrue(os.path.isfile(path))
            df = pd.read_csv(path)
            self.assertGreaterEqual(len(df), len(self.mixed_data))

    def test_save_csv_multi_file(self):
        with tempfile.TemporaryDirectory() as tmp:
            base = os.path.join(tmp, "out.csv")
            utils.save_results(base, self.mixed_data, filetype="csv", mode="multi_file")
            csv_files = [f for f in os.listdir(tmp) if f.endswith(".csv")]
            self.assertEqual(len(csv_files), len(self.mixed_data))

    def test_save_excel_single_sheet(self):
        with tempfile.TemporaryDirectory() as tmp:
            path = os.path.join(tmp, "out.xlsx")
            utils.save_results(path, self.mixed_data, filetype="excel", mode="single_sheet")
            self.assertTrue(os.path.isfile(path))
            df = pd.read_excel(path)
            self.assertGreaterEqual(len(df), len(self.mixed_data))

    def test_save_excel_multi_sheet(self):
        with tempfile.TemporaryDirectory() as tmp:
            path = os.path.join(tmp, "out.xlsx")
            utils.save_results(path, self.mixed_data, filetype="excel", mode="multi_sheet")
            self.assertTrue(os.path.isfile(path))
            xl = pd.ExcelFile(path)
            self.assertEqual(len(xl.sheet_names), len(self.mixed_data))

    def test_save_excel_multi_file(self):
        with tempfile.TemporaryDirectory() as tmp:
            base = os.path.join(tmp, "out.xlsx")
            utils.save_results(base, self.mixed_data, filetype="excel", mode="multi_file")
            xlsx_files = [f for f in os.listdir(tmp) if f.endswith(".xlsx")]
            self.assertEqual(len(xlsx_files), len(self.mixed_data))

    def test_save_with_descriptions(self):
        descriptions = [f"Item {i}" for i in range(len(self.mixed_data))]
        with tempfile.TemporaryDirectory() as tmp:
            path = os.path.join(tmp, "desc.csv")
            utils.save_results(path, self.mixed_data, filetype="csv", mode="single_sheet", descriptions=descriptions)
            df = pd.read_csv(path)
            self.assertIn("Description", df.columns)
            self.assertGreaterEqual(df["Description"].nunique(), len(self.mixed_data) // 2)  # flexible match

    def test_description_mismatch_raises(self):
        bad_descriptions = ["only one"]
        with self.assertRaises(ValueError):
            utils.save_results("test.csv", self.mixed_data, filetype="csv", descriptions=bad_descriptions)

    def test_invalid_mode_for_csv_raises(self):
        with self.assertRaises(ValueError):
            utils.save_results("bad.csv", self.mixed_data, filetype="csv", mode="multi_sheet")

    def test_invalid_filetype_raises(self):
        with self.assertRaises(ValueError):
            utils.save_results("bad.txt", self.mixed_data, filetype="txt")

    def test_invalid_mode_excel_raises(self):
        with self.assertRaises(ValueError):
            utils.save_results("bad.xlsx", self.mixed_data, filetype="excel", mode="nonsense_mode")



from VESIcal import sample_class, batchfile, calculate_classes

import pandas as pd
import numpy as np
import os

def flatten(obj):
    """
    Recursively converts a wide range of nested or mixed Python objects into a 
    list of pandas DataFrames, each representing one logical data unit.

    Supported input types include:
    - Scalars (int, float, str, and NumPy scalars like np.float64)
    - dicts and lists/tuples of dicts
    - pandas.DataFrame and pandas.Series
    - VESIcal objects with data access methods:
        - Calculate: uses `.result`
        - BatchFile: uses `.get_data()`
        - Sample: uses `.get_composition()`

    Parameters
    ----------
    obj : Any
        A single object or container of objects (e.g. list of dicts, list of Samples)
        to be flattened into one or more DataFrames.

    Returns
    -------
    list of pandas.DataFrame
        A list where each item is a flattened DataFrame representation of a top-level
        object in the input. If `obj` is not a list or tuple, the result is a single-item list.

    Raises
    ------
    TypeError
        If an unsupported object type is encountered during flattening.

    Examples
    --------
    >>> flatten({"A": 1, "B": 2})
    [   A  B
     0  1  2]

    >>> flatten([Sample(...), Calculate(...), {"x": 3}])
    [ DataFrame from Sample,
      DataFrame from Calculate,
      DataFrame from dict ]
    """

    def to_dataframe(item):
        if isinstance(item, calculate_classes.Calculate):
            return to_dataframe(item.result)

        elif isinstance(item, batchfile.BatchFile):
            return item.get_data().reset_index(drop=True)

        elif isinstance(item, sample_class.Sample):
            return to_dataframe(item.get_composition())

        elif isinstance(item, pd.DataFrame):
            return item.reset_index(drop=True)

        elif isinstance(item, pd.Series):
            return pd.DataFrame([item.to_dict()])

        elif isinstance(item, dict):
            return pd.DataFrame([item])

        elif isinstance(item, (list, tuple)):
            if all(isinstance(x, dict) for x in item):
                return pd.DataFrame(item)
            else:
                return pd.DataFrame([{"value": x} for x in item])

        elif isinstance(item, (int, float, str, np.generic)):
            return pd.DataFrame([{"value": item}])

        else:
            raise TypeError(f"Unsupported type in flatten: {type(item)}")

    if isinstance(obj, (list, tuple)):
        return [to_dataframe(item) for item in obj]
    return [to_dataframe(obj)]

def save_results(filename, obj, filetype="csv", mode="single_sheet", descriptions=None):
    """
    Save one or more data objects to CSV or Excel files with flexible formatting options.

    This function supports a wide range of input types, including scalars, dictionaries,
    pandas objects, and custom scientific objects (e.g., VESIcal Calculate, Sample, and BatchFile).
    It flattens each object into a pandas DataFrame before saving, and offers options for
    saving to a single file, multiple sheets, or separate files. Metadata can optionally be
    added to track the origin or meaning of each data entry.

    Parameters
    ----------
    filename : str
        Base filename to use for saving. The appropriate file extension will be added or replaced
        based on the selected `filetype`. When `mode='multi_file'`, index-based suffixes are appended
        automatically (e.g., `output_1.csv`, `output_2.csv`).

    obj : object or list/tuple of objects
        The data to save. Each object must be one of the supported types:
        - `pandas.DataFrame`, `pandas.Series`
        - `dict`, `list` of dicts
        - `int`, `float`, `str`, or any `numpy.generic` scalar (e.g., `np.float64`)
        - `Calculate` (accesses `.result`)
        - `Sample` (accesses `.get_composition()`)
        - `BatchFile` (accesses `.get_data()`)

    filetype : str, optional
        Output file format. Options are:
        - `"csv"`: Save as CSV file(s)
        - `"excel"`: Save as Excel file(s) (`.xlsx`)
        Default is `"csv"`.

    mode : str, optional
        Controls how multiple data objects are written. Options are:
        - `"single_sheet"`: All data combined into a single file and sheet (CSV or Excel)
        - `"multi_sheet"`: One Excel file with each item on its own sheet (Excel only)
        - `"multi_file"`: Each item saved to its own file (suffixes are added automatically)
        Default is `"single_sheet"`.

    descriptions : list of str, optional
        Optional metadata labels to add to each DataFrame as a `"Description"` column.
        Must match the number of top-level items in `obj` if `obj` is a list or tuple.
        Ignored if `obj` is a single object.

    Raises
    ------
    ValueError
        If `mode` is incompatible with the chosen `filetype`, or if the number of descriptions
        does not match the number of objects being saved.

    Examples
    --------
    >>> save_results("output.csv", [df1, df2], filetype="csv", mode="multi_file")

    >>> save_results("summary.xlsx", [sample1, calc], filetype="excel",
    ...              mode="multi_sheet", descriptions=["Basalt Sample", "Solubility Result"])
    """

    dfs = flatten(obj)

    if descriptions:
        if len(descriptions) != len(dfs):
            raise ValueError("Length of descriptions must match number of items.")
        for df, desc in zip(dfs, descriptions):
            df.insert(0, "Description", desc)

    if filetype == "csv":
        if mode == "single_sheet":
            combined = pd.concat(dfs, ignore_index=True)
            outname = filename if filename.endswith(".csv") else filename + ".csv"
            combined.to_csv(outname, index=False)
        elif mode == "multi_file":
            for i, df in enumerate(dfs):
                outname = f"{os.path.splitext(filename)[0]}_{i+1}.csv"
                df.to_csv(outname, index=False)
        else:
            raise ValueError("multi_sheet mode not supported for CSV. Use 'single_sheet' or 'multi_file'.")

    elif filetype == "excel":
        outname = filename if filename.endswith(".xlsx") else filename + ".xlsx"

        if mode == "single_sheet":
            combined = pd.concat(dfs, ignore_index=True)
            with pd.ExcelWriter(outname) as writer:
                combined.to_excel(writer, index=False, sheet_name="Sheet1")

        elif mode == "multi_sheet":
            with pd.ExcelWriter(outname) as writer:
                for i, df in enumerate(dfs):
                    sheetname = f"Item_{i+1}"
                    df.to_excel(writer, index=False, sheet_name=sheetname)

        elif mode == "multi_file":
            for i, df in enumerate(dfs):
                fname = f"{os.path.splitext(filename)[0]}_{i+1}.xlsx"
                df.to_excel(fname, index=False)
        else:
            raise ValueError("Invalid mode for Excel export. Choose 'single_sheet', 'multi_sheet', or 'multi_file'.")

    else:
        raise ValueError("Unsupported filetype. Choose 'csv' or 'excel'.")



import unittest
import VESIcal as v
import numpy as np
import pathlib
import pandas as pd
from pandas.testing import assert_frame_equal


# class TestSplitWeight(unittest.TestCase):
#     """
#     A class to define a method for asserting two pandas dataframes are equivalent.
#     From: https://stackoverflow.com/questions/38839402/how-to-use-assert-frame-equal-in-unittest
#     """
#     def assertDataframeEqual(self, a, b, msg):
#         try:
#             pd_testing.assert_frame_equal(a, b)
#         except AssertionError as e:
#             raise self.failureException(msg) from e

#     def setUp(self):
#         self.addTypeEqualityFunc(pd.DataFrame, self.assertDataframeEqual)

#     def test_allZero(self):
#         self.assertEqual(pd.DataFrame([0,0,0,0]), pd.DataFrame([0,0,0,0]))

def print_msg_box(msg, indent=1, width=None, title=None):
    """Print message-box with optional title."""
    lines = msg.split('\n')
    space = " " * indent
    if not width:
        width = max(map(len, lines))
    box = f'╔{"═" * (width + indent * 2)}╗\n'  # upper_border
    if title:
        box += f'║{space}{title:<{width}}{space}║\n'  # title
        box += f'║{space}{"-" * len(title):<{width}}{space}║\n'  # underscore
    box += ''.join([f'║{space}{line:<{width}}{space}║\n' for line in lines])
    box += f'╚{"═" * (width + indent * 2)}╝'  # lower_border
    print("\n")
    print(box)

print_msg_box("test_mansucript.py")

# Allow unittest to find the file
TEST_FILE = pathlib.Path(__file__).parent.joinpath("ManuscriptBatchFile.xlsx")
PICKLE_DISSOLVED = pathlib.Path(__file__).parent.joinpath("manuscript_dissolved.p")
PICKLE_EQFLUID = pathlib.Path(__file__).parent.joinpath("manuscript_eqfluid.p")
PICKLE_EQFLUID_WTEMPS = pathlib.Path(__file__).parent.joinpath("manuscript_eqfluid_wTemps.p")
PICKLE_EQFLUID_WT = pathlib.Path(__file__).parent.joinpath("manuscript_eqfluid_wt.p")
PICKLE_EQFLUID_MOL = pathlib.Path(__file__).parent.joinpath("manuscript_eqfluid_mol.p")
PICKLE_SATPS = pathlib.Path(__file__).parent.joinpath("manuscript_satPs.p")
PICKLE_SATPS_WTEMPS = pathlib.Path(__file__).parent.joinpath("manuscript_satPs_wTemps.p")
PICKLE_ISOBARS = pathlib.Path(__file__).parent.joinpath("manuscript_isobars.p")
PICKLE_ISOPLETHS = pathlib.Path(__file__).parent.joinpath("manuscript_isopleths.p")
PICKLE_CLOSED_DF = pathlib.Path(__file__).parent.joinpath("manuscript_closed.p")
PICKLE_OPEN_DF = pathlib.Path(__file__).parent.joinpath("manuscript_open_df.p")
PICKLE_HALF_DF = pathlib.Path(__file__).parent.joinpath("manuscript_half_df.p")
PICKLE_EXSOLVED_DF = pathlib.Path(__file__).parent.joinpath("manuscript_exsolved_df.p")
PICKLE_START2000_DF = pathlib.Path(__file__).parent.joinpath("manuscript_start2000_df.p")    

class TestManuscriptCalculations(unittest.TestCase):
    """
    Some of the "correct" values hard coded here are not exactly the same as those in the
    manuscript. However, they are very very close and likely are different because of changes
    to normalization routines made after publication. I've noted original values in commented
    text below where they differ from currently accepted values.

    Current values last updated March 2024 by Kayla Iacovino. 
    Some degassing paths updated March 2024 by Simon Matthews (see comments below).
    """
    def setUp(self):
        # import any pickled objects
        self.pickle_dissolved = pd.read_pickle(PICKLE_DISSOLVED)
        self.pickle_eqfluid = pd.read_pickle(PICKLE_EQFLUID)
        self.pickle_eqfluid_wTemps = pd.read_pickle(PICKLE_EQFLUID_WTEMPS)
        self.pickle_eqfluid_wt = pd.read_pickle(PICKLE_EQFLUID_WT)
        self.pickle_eqfluid_mol = pd.read_pickle(PICKLE_EQFLUID_MOL)
        self.pickle_satPs = pd.read_pickle(PICKLE_SATPS)
        self.pickle_satPs_wTemps = pd.read_pickle(PICKLE_SATPS_WTEMPS)
        self.pickle_isobars = pd.read_pickle(PICKLE_ISOBARS)
        self.pickle_isopleths = pd.read_pickle(PICKLE_ISOPLETHS)
        self.pickle_closed_df = pd.read_pickle(PICKLE_CLOSED_DF)
        self.pickle_open_df = pd.read_pickle(PICKLE_OPEN_DF)
        self.pickle_half_df = pd.read_pickle(PICKLE_HALF_DF)
        self.pickle_exsolved_df = pd.read_pickle(PICKLE_EXSOLVED_DF)
        self.pickle_start2000_df = pd.read_pickle(PICKLE_START2000_DF)

        # example data file
        self.myfile = v.BatchFile(TEST_FILE)

        # mysample used in ms
        self.mysample = v.Sample({'SiO2':  77.3,
                                  'TiO2':   0.08, 
                                  'Al2O3': 12.6,
                                  'Fe2O3':  0.207,
                                  'Cr2O3':  0.0,
                                  'FeO':    0.473,
                                  'MnO':    0.0,
                                  'MgO':    0.03,
                                  'NiO':    0.0,
                                  'CoO':    0.0,
                                  'CaO':    0.43,
                                  'Na2O':   3.98,
                                  'K2O':    4.88,
                                  'P2O5':   0.0,
                                  'H2O':    6.5,
                                  'CO2':    0.05})
        
        # values are same as in ms unless noted!
        self.mysample_satP = {'SaturationP_bars': 2960.0,  # changed from 2960 in ms due to rounding
                              'FluidMass_grams': 0.0018160337487088,
                              'FluidProportion_wt': 0.0018160337487087978,
                              'XH2O_fl': 0.838064480487942,
                              'XCO2_fl': 0.161935519512058}

        # mybasalt sample used in ms
        self.mybasalt = v.Sample({'SiO2': 47, 
                    'TiO2': 1.01,  
                    'Al2O3': 17.46,
                    'Fe2O3': 0.89,
                    'FeO': 7.18,
                    'MgO': 7.63,
                    'CaO': 12.44,
                    'Na2O': 2.65,
                    'K2O': 0.03,
                    'P2O5': 0.08, 
                    'CO2': 0.1})
        
        # normalize in three ways
        self.mybasalt_std = self.mybasalt.get_composition(normalization="standard",
                                                          asSampleClass=True)
        self.mybasalt_add = self.mybasalt.get_composition(normalization="additionalvolatiles",
                                                          asSampleClass=True)
        self.mybasalt_fix = self.mybasalt.get_composition(normalization="fixedvolatiles",
                                                          asSampleClass=True)
        
        self.manuscript_vals_norm_test = [1848.03, 1906.55, 1848.27, 1848.26]
        """
        Actual values from manuscript:
        Raw Saturation Pressure = 1848.031831425599 bars
        standard Saturation Pressure = 1906.5453789627868 bars
        additionalvolatiles Saturation Pressure = 1848.2673972122493 bars
        fixedvolatiles Saturation Pressure = 1848.2611364359402 bars
        """
        
        self.sample_10star = v.Sample({'SiO2':      47.9600,
                                        'TiO2':     0.7800,
                                        'Al2O3':    18.7700,
                                        'Fe2O3':    0.0000,
                                        'Cr2O3':    0.0000,
                                        'FeO':      10.9200,
                                        'MnO':      0.1500,
                                        'MgO':      6.8600,
                                        'NiO':      0.0000,
                                        'CoO':      0.0000,
                                        'CaO':      12.2300,
                                        'Na2O':     1.9500,
                                        'K2O':      0.2100,
                                        'P2O5':     0.1700,
                                        'H2O':      4.5000,
                                        'CO2':      0.0479 })
        
        self.tenstar_verbose_diss_vol = {'H2O_liq': 2.69337030386574,
                                        'CO2_liq': 0.0638470697446765,
                                        'XH2O_fl': 0.500053772677827,
                                        'XCO2_fl': 0.499946227322173,
                                        'FluidProportion_wt': 0.19679860317100492}
        """
        Actual values from manuscript:
        {'H2O_liq': 2.69352739399806,
        'CO2_liq': 0.0638439414375309,
        'XH2O_fl': 0.500092686493868,
        'XCO2_fl': 0.499907313506132,
        'FluidProportion_wt': 0.18407321260435108}
        """

        self.tenstar_eqfluid = {'CO2': 0.0052346121751066,
                                'H2O': 0.994765387824893}
        """
        Actual values from manuscript:
        {'CO2': 0.00528661429366132,
         'H2O': 0.994713385706339}
        """
    
    def test_degassing_paths(self):
        print_msg_box("TestManuscript \ndegassing_paths")
        """Calculate open, closed, and closed + 2 wt% initial vapor"""
        closed_df = v.calculate_degassing_path(sample=self.sample_10star, temperature=1200).result
        assert_frame_equal(closed_df, self.pickle_closed_df, atol=1e-4)

        open_df = v.calculate_degassing_path(sample=self.sample_10star, temperature=1200,
                                             fractionate_vapor=1.0).result
        assert_frame_equal(open_df, self.pickle_open_df, atol=1e-4)

        half_df = v.calculate_degassing_path(sample=self.sample_10star, temperature=1200,
                                             fractionate_vapor=0.5).result
        assert_frame_equal(half_df, self.pickle_half_df, atol=5e-4)

        exsolved_df = v.calculate_degassing_path(sample=self.sample_10star, temperature=1200,
                                                 init_vapor=2.0).result
        assert_frame_equal(exsolved_df, self.pickle_exsolved_df, atol=1e-4)

        """Calculate closed-system degassing starting from a pressure of 2000 bars"""
        start2000_df = v.calculate_degassing_path(sample=self.sample_10star, temperature=1200,
                                                  pressure=2000.0).result
        assert_frame_equal(start2000_df, self.pickle_start2000_df, atol=1e-4)
    
    def test_dissolved_batch(self):
        print_msg_box("TestManuscript \ndissolved_batch")
        result = self.myfile.calculate_dissolved_volatiles(temperature=900.0, pressure=2000.0,
                                                           X_fluid=1, print_status=True)
        assert_frame_equal(result, self.pickle_dissolved, atol=1e-4)
    
    def test_dissolved_volatiles_sample_10star(self):
        print_msg_box("TestManuscript \ndissolved_volatiles_sample_10star")
        result = v.calculate_dissolved_volatiles(sample=self.sample_10star, temperature=900.0,
                                                 pressure=2000.0, X_fluid=0.5, verbose=True).result
        params = list(result.keys())
        for param in params:
            self.assertAlmostEqual(result[param], self.tenstar_verbose_diss_vol[param], places=4)
    
    def test_eqfluid_batch(self):
        print_msg_box("TestManuscript \neqfluid_batch")
        result = self.myfile.calculate_equilibrium_fluid_comp(temperature=900.0, pressure=1000.0)
        assert_frame_equal(result, self.pickle_eqfluid, atol=1e-4)
    
    def test_eqfluid_batch_wTemps(self):
        print_msg_box("TestManuscript \neqfluid_batch_wTemps")
        result = self.myfile.calculate_equilibrium_fluid_comp(temperature='Temp', pressure='Press')
        assert_frame_equal(result, self.pickle_eqfluid_wTemps, atol=1e-4)
    
    def test_eqfluid_molfrac_to_wt(self):
        print_msg_box("TestManuscript \neqfluid_molfrac_to_wt")
        result = v.fluid_molfrac_to_wt(self.pickle_eqfluid)
        assert_frame_equal(result, self.pickle_eqfluid_wt, atol=1e-4)
    
    def test_eqfluid_wt_to_molfrac(self):
        print_msg_box("TestManuscript \neqfluid_wt_to_molfrac")
        result = v.fluid_wt_to_molfrac(self.pickle_eqfluid_wt)
        assert_frame_equal(result, self.pickle_eqfluid_mol, atol=1e-4)
    
    def test_equilibrium_fluid_sample_10star(self):
        print_msg_box("TestManuscript \nequilibrium_fluid_sample_10star")
        result = v.calculate_equilibrium_fluid_comp(sample=self.sample_10star, temperature=900.0,
                                                    pressure=100.0).result
        self.assertAlmostEqual(result['H2O'], self.tenstar_eqfluid['H2O'], places=4)
        self.assertAlmostEqual(result['CO2'], self.tenstar_eqfluid['CO2'], places=4)
    
    def test_isobars_and_isopleths(self):
        print_msg_box("TestManuscript \nisobars_and_isopleths")
        result_isobars, result_isopleths = v.calculate_isobars_and_isopleths(
                                            sample=self.sample_10star, 
                                            temperature=1200.0,
                                            pressure_list=[1000.0, 2000.0, 3000.0],
                                            isopleth_list=[0.25,0.5,0.75]).result
        assert_frame_equal(result_isobars, self.pickle_isobars, atol=1e-4)
        assert_frame_equal(result_isopleths, self.pickle_isopleths, atol=1e-4)
    
    def test_manuscript_plots(self):
        """
        This simply tests that plots created in the manuscript don't throw an error.
        """
        print_msg_box("TestManuscript \ntest_manuscript_plots")
        v.calib_plot()
        v.plot(isobars=self.pickle_isobars, isopleths=self.pickle_isopleths)
        v.plot(degassing_paths=[self.pickle_open_df, self.pickle_half_df, self.pickle_closed_df,
                                self.pickle_exsolved_df],
               degassing_path_labels=["Open", "Half", "Closed", "Exsolved"])
        v.plot(degassing_paths=[self.pickle_exsolved_df, self.pickle_start2000_df],
               degassing_path_labels=["Exsolved", "2000 bars"])
        v.plot(isobars=self.pickle_isobars, 
               isopleths=self.pickle_isopleths, 
               degassing_paths=[self.pickle_open_df, self.pickle_closed_df], 
               degassing_path_labels=["Open System", "Closed System"])

    def test_satP_different_normalizations(self):
        print_msg_box("TestManuscript \nsatP_different_normalization")
        for basalt, msval in zip([self.mybasalt, self.mybasalt_std, self.mybasalt_add,
                                  self.mybasalt_fix], self.manuscript_vals_norm_test):
            result = v.calculate_saturation_pressure(sample=basalt, temperature=1200,
                                                     model="IaconoMarziano").result
            self.assertAlmostEqual(result, msval, places=2)
    
    def test_saturation_pressure_batch(self):
        print_msg_box("TestManuscript \nsaturation_pressure_batch")
        result = self.myfile.calculate_saturation_pressure(temperature=925.0)
        assert_frame_equal(result, self.pickle_satPs, atol=1e-4)
    
    def test_saturation_pressure_batch_wTemps(self):
        print_msg_box("TestManuscript \nsaturation_pressure_batch_wTemps")
        result = self.myfile.calculate_saturation_pressure(temperature="Temp")
        assert_frame_equal(result, self.pickle_satPs_wTemps, atol=1e-4)
    
    def test_saturation_pressure_mysample(self):
        print_msg_box("TestManuscript \nsaturation_pressure_mysample")
        result = v.calculate_saturation_pressure(sample=self.mysample, temperature=925.0,
                                                 verbose=True).result
        params = list(result.keys())
        for param in params:
            self.assertAlmostEqual(result[param], self.mysample_satP[param], places=4)



import unittest
import VESIcal as v
import pandas as pd
import pathlib

# Allow unittest to find the file
TEST_FILE = pathlib.Path(__file__).parent.joinpath("ImportTest.xlsx")    

class TestCreateBatchFileFromDataFrame(unittest.TestCase):
    def assertDataframeEqual(self, a, b, msg):
        """
        Creates a new type of unittest to assert that pd DataFrames
        are equal, inheriting from pandas testing routine
        """
        try:
            pd._testing.assert_frame_equal(a, b)
        except AssertionError as e:
            raise self.failureException(msg) from e
        
    def setUp(self):
        self.addTypeEqualityFunc(pd.DataFrame, self.assertDataframeEqual)
        
        self.myfile = v.BatchFile(TEST_FILE)
        
        self.data_dict = {'SiO2':  [47.95, 47.95],
                     'TiO2':  [1.67, 1.67],
                     'Al2O3': [17.32, 17.32],
                     'FeO':   [10.24, 10.24],
                     'Fe2O3': [0.1, 0.1],
                     'MgO':   [5.76, 5.76],
                     'CaO':   [10.93, 10.93],
                     'Na2O':  [3.45, 3.45],
                     'K2O':   [1.99, 1.99],
                     'P2O5':  [0.51, 0.51],
                     'MnO':   [0.1, 0.1],
                     'H2O':   [2.0, 2.0],
                     'CO2':   [0.1, 0.1],
                     'Notes': ['Normal sample', 'Duplicate sample']}
        
        self.df = pd.DataFrame(self.data_dict, 
                               index=['test_samp',
                                      'test_samp-duplicate-1'])
        self.bf = v.BatchFile_from_DataFrame(self.df)
    
    def test_BatchFileFromDataFrame(self):
        self.assertEqual(self.bf.get_data(), self.myfile.get_data(), 
                         'BatchFiles are different')
        
class TestGetData(unittest.TestCase):
    def test_function_processes_all_oxides(self):
        """Verify conversion functions can process every oxide."""
        sample_all_oxides = {ox: i+1 for i, ox in enumerate(v.core.oxides)}
        df_all_oxides = pd.DataFrame(sample_all_oxides, index=[0])
        bf_all_oxides = v.BatchFile_from_DataFrame(df_all_oxides)
        
        for unitname in ['wtpt_oxides', 'mol_oxides', 'mol_cations',
                         'mol_singleO']:
            result = bf_all_oxides.get_data(units=unitname)
            
            # Verify we got a result
            self.assertIsNotNone(result)
            self.assertGreater(len(result), 0)        
        


import unittest
from VESIcal import core
from VESIcal import sample_class
import pandas as pd
    
class TestOxideCationConversions(unittest.TestCase):
    def setUp(self):
        self.oxide_data_dict_keys = [
            'mass',
            'cation',
            'cation_num',
            'oxygen_num',
            'cation_charge',
            'cation_mass',
            'groups', 
        ]
        
    def test_all_oxide_cation_mappings_complete(self):
        """Verify all oxides have cations and all mappings are consistent."""
                
        # Test 1: All oxides and cations in conversion dicts
        for oxide in core.oxides:
            with self.subTest(oxide=oxide):
                # Check oxide->cation mapping exists
                self.assertIn(oxide, core.oxides_to_cations,
                            f"{oxide} missing from oxides_to_cations")
        
        for cation in core.cations:
            with self.subTest(cation=cation):
                self.assertIn(cation, core.cations_to_oxides,
                              f"{cation} missing from cations_to_oxides")
        
        # Test 2: All oxides have all values defined in oxide_data dict
        for oxide in core.oxides:
            with self.subTest(oxide=oxide):
                for k in core.oxide_data[oxide].keys():
                    self.assertIn(k,
                              self.oxide_data_dict_keys,
                              f"{oxide} has incomplete data in oxide_data dict")
        
        # Test 3: All oxides in property dictionaries 
        # Check oxide->cation mapping exists
        for oxide in core.oxides:
            with self.subTest(oxide=oxide):
                self.assertIn(oxide, core.oxideMass,
                            f"{oxide} missing from oxideMass")
                
                self.assertIn(oxide, core.CationNum,
                            f"{oxide} missing from CationNum")
                
                self.assertIn(oxide, core.OxygenNum,
                            f"{oxide} missing from OxygenNum")
                
                self.assertIn(oxide, core.CationCharge,
                            f"{oxide} missing from CationCharge")
                
                self.assertIn(oxide, core.CationMass,
                            f"{oxide} missing from CationMass")


"""
Single Source of Truth pattern for oxide-cation compositional factors.
All derived data structures are computed from a single master dictionary.
"""

# Master data structure - single source of truth
oxide_data = {
    'SiO2': {
        'mass': 60.083,
        'cation': 'Si',
        'cation_num': 1,
        'oxygen_num': 2,
        'cation_charge': 4,
        'cation_mass': 28.085,
        'groups': ['magmasat', 'anhydrous']
    },
    'TiO2': {
        'mass': 79.867,
        'cation': 'Ti',
        'cation_num': 1,
        'oxygen_num': 2,
        'cation_charge': 4,
        'cation_mass': 47.867,
        'groups': ['magmasat', 'anhydrous']
    },
    'Al2O3': {
        'mass': 101.961,
        'cation': 'Al',
        'cation_num': 2,
        'oxygen_num': 3,
        'cation_charge': 3,
        'cation_mass': 26.982,
        'groups': ['magmasat', 'anhydrous']
    },
    'Cr2O3': {
        'mass': 151.992,
        'cation': 'Cr',
        'cation_num': 2,
        'oxygen_num': 3,
        'cation_charge': 3,
        'cation_mass': 51.996,
        'groups': ['magmasat', 'anhydrous']
    },
    'FeO': {
        'mass': 71.844,
        'cation': 'Fe',
        'cation_num': 1,
        'oxygen_num': 1,
        'cation_charge': 2,
        'cation_mass': 55.845,
        'groups': ['magmasat', 'anhydrous']
    },
    'Fe2O3': {
        'mass': 159.687,
        'cation': 'Fe3',
        'cation_num': 2,
        'oxygen_num': 3,
        'cation_charge': 3,
        'cation_mass': 55.845,
        'groups': ['magmasat', 'anhydrous']
    },
    'MnO': {
        'mass': 70.937,
        'cation': 'Mn',
        'cation_num': 1,
        'oxygen_num': 1,
        'cation_charge': 2,
        'cation_mass': 54.938,
        'groups': ['magmasat', 'anhydrous']
    },
    'MgO': {
        'mass': 40.304,
        'cation': 'Mg',
        'cation_num': 1,
        'oxygen_num': 1,
        'cation_charge': 2,
        'cation_mass': 24.305,
        'groups': ['magmasat', 'anhydrous']
    },
    'CaO': {
        'mass': 56.077,
        'cation': 'Ca',
        'cation_num': 1,
        'oxygen_num': 1,
        'cation_charge': 2,
        'cation_mass': 40.078,
        'groups': ['magmasat', 'anhydrous']
    },
    'Na2O': {
        'mass': 61.979,
        'cation': 'Na',
        'cation_num': 2,
        'oxygen_num': 1,
        'cation_charge': 1,
        'cation_mass': 22.990,
        'groups': ['magmasat', 'anhydrous']
    },
    'K2O': {
        'mass': 94.195,
        'cation': 'K',
        'cation_num': 2,
        'oxygen_num': 1,
        'cation_charge': 1,
        'cation_mass': 39.098,
        'groups': ['magmasat', 'anhydrous']
    },
    'P2O5': {
        'mass': 141.943,
        'cation': 'P',
        'cation_num': 2,
        'oxygen_num': 5,
        'cation_charge': 5,
        'cation_mass': 30.974,
        'groups': ['magmasat', 'anhydrous']
    },
    'NiO': {
        'mass': 74.692,
        'cation': 'Ni',
        'cation_num': 1,
        'oxygen_num': 1,
        'cation_charge': 2,
        'cation_mass': 58.693,
        'groups': ['magmasat', 'anhydrous']
    },
    'CoO': {
        'mass': 44.01,
        'cation': 'Co',
        'cation_num': 1,
        'oxygen_num': 1,
        'cation_charge': 2,
        'cation_mass': 28.01,
        'groups': ['magmasat', 'anhydrous']
    },
    'H2O': {
        'mass': 18.02,
        'cation': 'H',
        'cation_num': 2,
        'oxygen_num': 1,
        'cation_charge': 1,
        'cation_mass': 1.01,
        'groups': ['magmasat', 'volatile']
    },
    'CO2': {
        'mass': 44.01,
        'cation': 'C',
        'cation_num': 1,
        'oxygen_num': 2,
        'cation_charge': 4,
        'cation_mass': 12.011,
        'groups': ['magmasat', 'volatile']
    },
    'F2O': {
        'mass': 37.997,
        'cation': 'F',
        'cation_num': 2,
        'oxygen_num': 1,
        'cation_charge': 1,
        'cation_mass': 18.998,
        'groups': ['anhydrous']
    }
}

# Derived data structures - all computed from oxide_data
# Lists
oxides = list(oxide_data.keys())
cations = list(set(data['cation'] for data in oxide_data.values()))
magmasat_oxides = [oxide for oxide, data in oxide_data.items() if 'magmasat' in data['groups']]
anhydrous_oxides = [oxide for oxide, data in oxide_data.items() if 'anhydrous' in data['groups']]
volatiles = [oxide for oxide, data in oxide_data.items() if 'volatile' in data['groups']]

# Property dictionaries
oxideMass = {oxide: data['mass'] for oxide, data in oxide_data.items()}
CationNum = {oxide: data['cation_num'] for oxide, data in oxide_data.items()}
OxygenNum = {oxide: data['oxygen_num'] for oxide, data in oxide_data.items()}
CationCharge = {oxide: data['cation_charge'] for oxide, data in oxide_data.items()}
CationMass = {oxide: data['cation_mass'] for oxide, data in oxide_data.items()}

# Conversion mappings
oxides_to_cations = {oxide: data['cation'] for oxide, data in oxide_data.items()}
cations_to_oxides = {data['cation']: oxide for oxide, data in oxide_data.items()}


# ---------- DATA TRANSFORMATION FOR PANDAS DATAFRAMES --------- #
def fluid_molfrac_to_wt(data, H2O_colname='XH2O_fl_VESIcal', CO2_colname='XCO2_fl_VESIcal'):
    """
    Takes in a pandas dataframe object and converts only the fluid composition from mole fraction
    to wt%, leaving the melt composition in tact. The user must specify the names of the
    XH2O_fl and XCO2_fl columns.

    Parameters
    ----------
    data: pandas DataFrame
        Sample composition(s) containing columns for H2O and CO2 concentrations in the fluid.

    H2O_colname: str
        OPTIONAL. The default value is 'XH2O_fl', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the H2O
        concentration in the fluid, in mol fraction.

    CO2_colname: str
        OPTIONAL. The default value is 'XCO2_fl', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the CO2
        concentration in the fluid, in mol fraction.

    Returns
    -------
    pandas DataFrame
        Original data passed plus newly calculated values are returned.
    """
    convData = data.copy()

    MPO_H2O_list = []
    MPO_CO2_list = []
    for index, row in convData.iterrows():
        MPO_H2O_list.append(row[H2O_colname] * oxideMass["H2O"])
        MPO_CO2_list.append(row[CO2_colname] * oxideMass["CO2"])

    convData["MPO_H2O"] = MPO_H2O_list
    convData["MPO_CO2"] = MPO_CO2_list
    convData["H2O_fl_wt"] = 100 * convData["MPO_H2O"] / (convData["MPO_H2O"] + convData["MPO_CO2"])
    convData["CO2_fl_wt"] = 100 * convData["MPO_CO2"] / (convData["MPO_H2O"] + convData["MPO_CO2"])

    del convData["MPO_H2O"]
    del convData["MPO_CO2"]

    return convData


def fluid_wt_to_molfrac(data, H2O_colname='H2O_fl_wt', CO2_colname='CO2_fl_wt'):
    """
    Takes in a pandas dataframe object and converts only the fluid composition from wt% to mole
    fraction, leaving the melt composition in tact. The user must specify the names of the
    H2O_fl_wt and CO2_fl_wt columns.

    Parameters
    ----------
    data: pandas DataFrame
        DataFrame containing columns for H2O and CO2 concentrations in the fluid.

    H2O_colname: str
        OPTIONAL. The default value is 'H2O_fl_wt', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the H2O
        concentration in the fluid, in wt%.

    CO2_colname: str
        OPTIONAL. The default value is 'CO2_fl_wt', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the CO2
        concentration in the fluid, in wt%.

    Returns
    -------
    pandas DataFrame
        Original data passed plus newly calculated values are returned.
    """
    convData = data.copy()

    MPO_H2O_list = []
    MPO_CO2_list = []
    for index, row in convData.iterrows():
        MPO_H2O_list.append(row[H2O_colname] / oxideMass["H2O"])
        MPO_CO2_list.append(row[CO2_colname] / oxideMass["CO2"])

    convData["MPO_H2O"] = MPO_H2O_list
    convData["MPO_CO2"] = MPO_CO2_list
    convData["XH2O_fl"] = convData["MPO_H2O"] / (convData["MPO_H2O"] + convData["MPO_CO2"])
    convData["XCO2_fl"] = convData["MPO_CO2"] / (convData["MPO_H2O"] + convData["MPO_CO2"])

    del convData["MPO_H2O"]
    del convData["MPO_CO2"]

    return convData


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class InputError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class SaturationError(Error):
    """Exception raised for errors thrown when a sample does not reach saturation.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="VESIcal",
    version="1.2.12",
    author="Kayla Iacovino, Simon Matthews, Penny Wieser",
    author_email="kaylaiacovino@gmail.com",
    description=("A generalized python library for calculating and plotting various things "
                 "related to mixed volatile (H2O-CO2) solubility in silicate melts."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kaylai/VESIcal",
    packages=setuptools.find_packages(),
    install_requires=[
            'pandas',
            'numpy',
            'matplotlib',
            'scipy',
            'sympy',
            'openpyxl'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
